(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: CorrSortXor.v,v 1.2 2006/07/17 23:29:15 monin Exp $ *)

Require Import List. 
Require Import RelProps.
Require Import Separable. 
Require Import LasagnaTerms. (* using xor_term *)
Require Import SortXor.


(* *)

Section sec_corr_sort.

Variable AS : sortable.

Notation A := (|AS|). 
Notation ueqA := (ueq AS). 
Notation ueqA_True_False := (ueq_True_False AS).
Notation ueqA_equal := (ueq_equal AS).
Notation ueqA_refl := (ueq_refl AS).

Notation ultA := (ult AS).
Notation ultA_True_False := (ult_True_False AS).
Notation ultA_ueqA_False := (ult_ueq_False AS).
Notation ultA_trans := (ult_trans AS).
Notation ultA_total := (ult_total AS).

Inductive equiv : forall b1, xor_term A b1 -> forall b2, xor_term A b2 -> Prop :=
  | refl_Z_equiv : forall b1 b2, equiv b1 (X_Zero b1) b2 (X_Zero b2)
  | congr_I_equiv : forall a, equiv true (X_ns true I a) true (X_ns true I a)
  | sym_equiv : forall bx x by y, equiv bx x by y -> equiv by y bx x
  | trans_equiv : forall bx x by y bz z, 
     equiv bx x by y -> equiv by y bz z -> equiv bx x bz z
  | congr_X_quiv : forall b1 b2 x1 x2 y1 y2, 
     equiv true x1 true x2 -> equiv true y1 true y2 -> 
     equiv b1 (X_Xor b1 x1 y1) b2 (X_Xor b2 x2 y2)
  | comm_equiv : forall b1 b2 x y, equiv b1 (X_Xor b1 x y) b2 (X_Xor b2 y x)
  | assoc_equiv : forall b1 b2 x y z,
      equiv b1 (X_Xor b1 (X_Xor true x y) z) b2 (X_Xor b2 x (X_Xor true y z))
  | neutral_equiv : forall b x, equiv b (X_Xor b x (X_Zero _)) true x
  | invol_equiv : forall b1 b2 x, equiv b1 (X_Xor b1 x x) b2 (X_Zero _).

Hint Resolve
  refl_Z_equiv congr_I_equiv congr_X_quiv neutral_equiv invol_equiv comm_equiv : 
  equiv_db.

Lemma refl_equiv : forall b x, equiv b x b x.
induction x; auto with equiv_db.
destruct b; case i. apply congr_I_equiv. 
Qed.

Hint Resolve refl_equiv : equiv_db.

Fixpoint sem b (l : list A) {struct l} : xor_term A b :=
   match l with
     | nil => X_Zero b
     | x :: l => X_Xor b (X_ns true I x) (sem true l)
   end.

Lemma refl_sem : forall l b1 b2, equiv b1 (sem b1 l) b2 (sem b2 l).
induction l as [ | a l ]; simpl; intros b1 b2.
   apply refl_Z_equiv. 
   apply congr_X_quiv. 
     apply congr_I_equiv. 
     apply IHl. 
Qed.

Theorem equiv_sem : forall b1 x1 b2 x2, equiv b1 x1 b2 x2 -> sort x1 = sort x2.
induction 1;  unfold sort in * |- * ; simpl. 
  reflexivity.
  reflexivity.
  symmetry. exact IHequiv.
  rewrite IHequiv1. exact IHequiv2.  
  rewrite IHequiv2. rewrite commut. rewrite IHequiv1. rewrite <- commut. reflexivity.
  apply commut. 
  reflexivity. 
  reflexivity. 
  apply fold_insert_invol.
Qed.


Theorem sem_sort : forall l, sorted l -> forall b, sort (sem b l) = l. 
induction l; unfold sort in * |- * ; simpl. 
  reflexivity. 
  intros sal b. rewrite IHl. 
    rewrite sorted1_insert; trivial.
    apply inv_sorted1 with a. assumption.
Qed.

Hint Resolve refl_sem :
  equiv_db.

Lemma corr_insert_1_elt : 
  forall a l, sorted l -> forall b1 b2,
  equiv b1 (sem b1 (insert AS a l)) b2 (X_Xor b2 (X_ns true I a) (sem true l)).
intros a. induction l as [ | d l]; simpl. 
    intros t. auto with equiv_db. 
    intros sdl b1 b2. case (ultA_total a d); simpl; intro c.
        auto with equiv_db. 
        rewrite c. apply trans_equiv with
 true (X_Xor true (X_Xor true (X_ns true I d) (X_ns true I d)) (sem true l)).
           apply trans_equiv with true (X_Xor true (sem true l) (X_Zero true)). 
              apply sym_equiv. apply trans_equiv with true (sem true l); auto with equiv_db. 
              apply trans_equiv with true (X_Xor true (X_Zero true) (sem true l));
                                       auto with equiv_db. 
                apply congr_X_quiv; auto with equiv_db. 
                  apply sym_equiv. apply invol_equiv. 
           apply assoc_equiv. 
        apply trans_equiv with
 true (X_Xor true (X_ns true I d) (X_Xor true (X_ns true I a) (sem true l))).
           apply congr_X_quiv. 
             apply congr_I_equiv. 
             apply IHl. apply inv_sorted1 with d; assumption.
           apply trans_equiv with 
 true (X_Xor true (X_Xor true (X_ns true I a) (X_ns true I d)) (sem true l)).
             apply sym_equiv. apply trans_equiv with
 true (X_Xor true (X_Xor true (X_ns true I d) (X_ns true I a)) (sem true l));
      auto with equiv_db.
                apply assoc_equiv. 
              apply assoc_equiv.
Qed.

Theorem sem_sort_equiv_gen : 
  forall x l, sorted l -> forall b1 b2,
  equiv b1 (sem b1 (fold_insert x l)) b2 (X_Xor b2 x (sem true l)).
intro x. elim x using xor_term_true_ind; clear x; simpl. 
  intros l sl b1 b2. apply sym_equiv. apply trans_equiv with
                true (X_Xor true (sem true l) (X_Zero true)).
    apply comm_equiv. 
    apply trans_equiv with true (sem true l); auto with equiv_db.

  intros a l sl b1 b2. apply corr_insert_1_elt. assumption. 

  intros x y Hx Hy l sl b1 b2. generalize (Hx (fold_insert y l)). intro Hrec. 
    apply trans_equiv with true (X_Xor true x (sem true (fold_insert y l))). 
       apply Hrec. apply fold_insert_is_sorted. assumption. 
       apply trans_equiv with true (X_Xor true x (X_Xor true y (sem true l))).
          apply congr_X_quiv. 
             apply refl_equiv. 
             apply Hy. assumption. 
          apply sym_equiv. apply assoc_equiv. 
Qed.

Theorem sem_sort_equiv_true : forall b x, equiv b (sem b (sort x)) true x.
unfold sort. intros b x. generalize (sem_sort_equiv_gen x nil I); simpl; intro e. 
apply trans_equiv with true (X_Xor true x (X_Zero true)). 
   apply e. 
   apply neutral_equiv. 
Qed.

Theorem sem_sort_equiv_false : forall b x, equiv b (sem b (sort x)) false x.
unfold sort. intros b x. 
elim x using xor_term_false_ind; clear x; simpl. 
  apply refl_Z_equiv. 
  intros x y. 
generalize (sem_sort_equiv_gen x (fold_insert y nil)). intro e. 
apply trans_equiv with true (X_Xor true x (sem true (fold_insert y nil))). 
   apply e. apply fold_insert_is_sorted. exact I.
   apply congr_X_quiv. 
     apply refl_equiv.
     exact (sem_sort_equiv_true true y). 
Qed.


End sec_corr_sort.
