(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: Lasagna.v,v 1.6 2006/03/08 13:35:24 monin Exp $ *)

Require Import Nat2.
Require Import Terms.
Require Import Empty.

Require Import LasagnaTerms.

Fixpoint lift_lasagna_x e1 e2 {struct e1} : lasagna_x e1 -> lasagna_x (e1 +e e2) :=
  match e1 return lasagna_x e1 -> lasagna_x (e1 +e e2) with
  | AZ => fun emp => match emp with end
  | ES o1 => map_xor (lift_lasagna_n o1 e2) false
  end
with lift_lasagna_n o1 e2 {struct o1} : lasagna_n o1 -> lasagna_n (o1 +o e2) :=
  match o1 return lasagna_n o1 -> lasagna_n (o1 +o e2) with
  | OS e1 => map_nx (lift_lasagna_x e1 e2) false
  end.

Definition lift_lasagna_cand_x b o1 e2 : 
  lasagna_cand_x o1 b -> lasagna_cand_x (o1 +o e2) b :=
  map_xor (lift_lasagna_n o1 e2) b.

Definition lift_lasagna_cand_n b e1 e2 :
  lasagna_cand_n e1 b -> lasagna_cand_n (e1 +e e2) b :=
  map_nx (lift_lasagna_x e1 e2) b.

(* ---------------------------------------------- *)
(* Ltac as an editor of complex convertible types *)
(*
Definition interactive_bin_nx :
  (forall A b, nx_term A true -> nx_term A true -> nx_term A b) -> 
  forall e1 e2 b, 
  lasagna_cand_n e1 true -> lasagna_cand_n e2 true -> lasagna_cand_n (max_ee e1 e2) b.
intros bin e1 e2 b l1 l2.
match goal with [|- ?g ?e ?b] => red; apply bin; change (g e true) end.
  unfold max_ee. apply lift_lasagna_cand_n. exact l1.
  rewrite sym_max_ee. unfold max_ee. apply lift_lasagna_cand_n. exact l2.
Defined.

Definition interactive_bin_xor :
  (forall A b, xor_term A true -> xor_term A true -> xor_term A b) -> 
  forall o1 o2 b, 
  lasagna_cand_x o1 true -> lasagna_cand_x o2 true -> lasagna_cand_x (max_oo o1 o2) b.
intros bin o1 o2 b l1 l2.
match goal with [|- ?g ?o ?b] => red; apply bin; change (g o true) end.
  unfold max_oo. apply lift_lasagna_cand_x. exact l1.
  rewrite sym_max_oo. unfold max_oo. apply lift_lasagna_cand_x. exact l2.
Defined.
*)
(* ---------------------------------------------- *)

Definition coerce_max_ee_comm : forall e1 e2, 
  lasagna_cand_n (max_ee e2 e1) true -> lasagna_cand_n (max_ee e1 e2) true.
intros e1 e2 l; rewrite sym_max_ee; trivial.
Defined.

Implicit Arguments coerce_max_ee_comm [e1 e2].

Definition bin_nx
  (bin : forall A b,  nx_term A true -> nx_term A true -> nx_term A b) e1 e2 b
  (l1 : lasagna_cand_n e1 true) (l2 : lasagna_cand_n e2 true) :
  lasagna_cand_n (max_ee e1 e2) b  :=
  bin (lasagna_x (max_ee e1 e2)) b
      (lift_lasagna_cand_n true e1 (e2 -e e1) l1)
      (coerce_max_ee_comm (lift_lasagna_cand_n true e2 (e1 -e e2) l2)).

Definition coerce_max_oo_comm : forall o1 o2, 
  lasagna_cand_x (max_oo o2 o1) true -> lasagna_cand_x (max_oo o1 o2) true.
intros o1 o2 l; rewrite sym_max_oo; trivial.
Defined.

Implicit Arguments coerce_max_oo_comm [o1 o2].

Definition bin_xor
  (bin : forall A b,  xor_term A true -> xor_term A true -> xor_term A b) o1 o2 b
  (l1 : lasagna_cand_x o1 true) (l2 : lasagna_cand_x o2 true) :
  lasagna_cand_x (max_oo o1 o2) b  :=
  bin (lasagna_n (max_oo o1 o2)) b
      (lift_lasagna_cand_x true o1 (o2 -o o1) l1)
      (coerce_max_oo_comm (lift_lasagna_cand_x true o2 (o1 -o o2) l2)).


Implicit Arguments bin_nx [e1 e2 b].
Implicit Arguments bin_xor [o1 o2 b].


(* ---------------------------------------------------------------------- *)

Definition inj_even_parity p : alte -> alt_of_parity p :=
    match p return alte -> alt_of_parity p with
    | Pev => fun e => e
    | Pod => OS 
    end.

Definition inj_odd_parity p : alto -> alt_of_parity p :=
    match p return alto -> alt_of_parity p with
    | Pev => ES
    | Pod => fun o => o
    end.

Fixpoint alt_allpar_of_term (t:term) : forall p, alt_of_parity p :=
  match t return forall p, alt_of_parity p with
  | Zero => fun p => inj_odd_parity p (OS AZ)
  | Xor x y => 
    let o1 := alt_allpar_of_term x Pod in
    let o2 := alt_allpar_of_term y Pod in
    fun p => inj_odd_parity p (max_oo o1 o2)

  | PC x => fun p => inj_even_parity p AZ
  | SC x => fun p => inj_even_parity p AZ
  | E x y => 
    let e1 := alt_allpar_of_term x Pev in
    let e2 := alt_allpar_of_term y Pev in
    fun p => inj_even_parity p (max_ee e1 e2)
  | Hash x y => 
    let e1 := alt_allpar_of_term x Pev in
    let e2 := alt_allpar_of_term y Pev in
    fun p => inj_even_parity p (max_ee e1 e2)
  end.

Definition kind_lasagna_cand_of_term (t:term) (p: parity) : Set :=
  match p with
  | Pev => lasagna_cand_n (alt_allpar_of_term t Pev) true
  | Pod => lasagna_cand_x (alt_allpar_of_term t Pod) true
  end.

Fixpoint lasagna_cand_of_term (t:term) :
  forall p, kind_lasagna_cand_of_term t p :=
  match t return forall p, kind_lasagna_cand_of_term t p with
  | Zero => 
      fun p => match p return kind_lasagna_cand_of_term Zero p with
      | Pev => NX_sumt (X_Zero false)
      | Pod => X_Zero true
      end
  | Xor t1 t2 => 
      let l1 := lasagna_cand_of_term t1 Pod in
      let l2 := lasagna_cand_of_term t2 Pod in
      fun p => match p return kind_lasagna_cand_of_term (Xor t1 t2) p with
      | Pev => NX_sumt (bin_xor X_Xor l1 l2)
      | Pod => bin_xor X_Xor l1 l2
      end

  | PC x => 
      fun p => match p return kind_lasagna_cand_of_term (PC x) p with
      | Pev => NX_PC true x
      | Pod => X_nst (NX_PC false x)
      end
  | SC x => 
      fun p => match p return kind_lasagna_cand_of_term (SC x) p with
      | Pev => NX_SC true x
      | Pod => X_nst (NX_SC false x)
      end
  | E t1 t2 => 
      let l1 := lasagna_cand_of_term t1 Pev in
      let l2 := lasagna_cand_of_term t2 Pev in
      fun p => match p return kind_lasagna_cand_of_term (E t1 t2) p with
      | Pev => bin_nx NX_E l1 l2
      | Pod => X_nst (bin_nx NX_E l1 l2)
      end
  | Hash t1 t2 => 
      let l1 := lasagna_cand_of_term t1 Pev in
      let l2 := lasagna_cand_of_term t2 Pev in
      fun p => match p return kind_lasagna_cand_of_term (Hash t1 t2) p with
      | Pev => bin_nx NX_Hash l1 l2
      | Pod => X_nst (bin_nx NX_Hash l1 l2)
      end
end.

(* *)

Definition parity_of_term t : parity :=
  match t with
  | Zero => Pev
  | Xor t1 t2 => Pev

  | PC x => Pod
  | SC x => Pod
  | E t1 t2 => Pod
  | Hash t1 t2 => Pod
end.


Definition alt_of_term t := alt_allpar_of_term t (parity_of_term t).

Definition lasagna_of_parity p : alt_of_parity p -> Set :=
  match p return alt_of_parity p -> Set with
  | Pev => lasagna_x
  | Pod => lasagna_n
  end.

Definition lasagna_of_term (t:term) :
  lasagna_of_parity (parity_of_term t) (alt_of_term t) :=
  match t return lasagna_of_parity (parity_of_term t) (alt_of_term t) with
  | Zero => X_Zero false
  | Xor t1 t2 => 
      let l1 := lasagna_cand_of_term t1 Pod in
      let l2 := lasagna_cand_of_term t2 Pod in
      bin_xor X_Xor l1 l2

  | PC x => NX_PC false x
  | SC x => NX_SC false x
  | E t1 t2 => 
      let l1 := lasagna_cand_of_term t1 Pev in
      let l2 := lasagna_cand_of_term t2 Pev in
      bin_nx NX_E l1 l2
  | Hash t1 t2 => 
      let l1 := lasagna_cand_of_term t1 Pev in
      let l2 := lasagna_cand_of_term t2 Pev in
      bin_nx NX_Hash l1 l2
  end.

(* OUF ! *)
(* ---------------------------------------------------------------------- *)

(*  
fonction inverse -> prouver completude et unicite
modulo le pb souleve par Judicael

*)

Fixpoint term_x (e:alte) : lasagna_x e -> term :=
  match e return lasagna_x e -> term with
  | AZ => fun em => match em with end
  | ES o => map_xor_term (term_n o) false
  end
with term_n (o:alto) : lasagna_n o -> term :=
  match o return lasagna_n o -> term with
  | OS e => map_nx_term (term_x e) false
  end.

(* A FAIRE : prouver des choses de ce genre 
Lemma aux_term_x_lasagna :
  forall o  (f: lasagna_n o -> term) t,
  lasagna_of_term  o f (lasagna_of_term t) = t.

Theorem term_lasagna_x : 
  forall t, term_x _ (lasagna_x_of_term t) = t.
*)


(* ---------------------------------------------------------------------- *)

(* interactive version
Definition lasagna_cand_of_term (t:term) : forall p, kind_lasagna_cand_of_term t p.
intro t; elim t; clear t.
  intro p ; case p.
    exact (NX_sumt (X_Zero false)). 
    exact (X_Zero true). 

  intros p pa; case pa; clear pa.
    exact (NX_PC true p). 
    exact (X_nst (NX_PC false p)). 

  intros s pa; case pa; clear pa.
    exact (NX_SC true s). 
    exact (X_nst (NX_SC false s)). 

  intros t1 l1 t2 l2 p. case p; clear p; simpl. 
    exact (bin_nx NX_E (l1 Pev) (l2 Pev)).
    exact (X_nst (bin_nx NX_E (l1 Pev) (l2 Pev))).

  intros t1 l1 t2 l2 p. case p; clear p; simpl. 
    exact (NX_sumt  (bin_xor X_Xor (l1 Pod) (l2 Pod))).
    exact (bin_xor X_Xor (l1 Pod) (l2 Pod)).

  intros t1 l1 t2 l2 p. case p; clear p; simpl. 
    exact (bin_nx NX_Hash (l1 Pev) (l2 Pev)).
    exact (X_nst (bin_nx NX_Hash (l1 Pev) (l2 Pev))).
Defined.
*)

(* On inverse le role de p, qui repr la couche suivante attendue
   et on la couche courante
Ex. Je suis un E t1 t2, donc impair, qui produira un nx_term.
Mon voisin du dessus peut tre 
  - un Xor => il m'a dit qu'il attendait un Pod au bout, ce que je suis
   => je lui fournis un X_ns true (NX_E false t1' t2')
      o ti' est rcupr de ti en disant qu'on attend un Pev au bout
      NB : NX_E false t1' t2' de type nx_term (...) false = lasagna_n
           X_ns (NX_E false t1' t2') de type xor_term (...) true
   Mon propre E incrmente :
                   t1' : nx_term (lasagna_x e1) true 
                   t2' : nx_term (lasagna_x e2) true
           ------------------------------------------------------------
           NX_E false t1' t2' : nx_term (lasagna_x (max e1 e2)) false =
                                lasagna_n (OS (max e1 e2))
     ---------------------------------------------------------------------------
     X_ns true (NX_E false t1' t2') : xor_term (lasagna_n (OS (max e1 e2))) true
   => o = OS (max e1 e2)

  - un E x y => il m'a dit qu'il attendait un Pev au bout, ce que je ne suis pas
   => je lui fournis un NX_E true t1' t2' de type nx_term (...) true
      o ti' est rcupr de ti en disant qu'on attend un Pev au bout
   Mon propre E n'incrmente rien :
                   t1' : nx_term (lasagna_x e1) true 
                   t2' : nx_term (lasagna_x e2) true
           --------------------------------------------------------
           NX_E true t1' t2' : nx_term (lasagna_x (max e1 e2)) true
   => e = max e1 e2
 *)

(* ---------------------------------------------------------------------- *)
