(* AUTEURS / AUTHORS :                          *)
(* Judical Courant,   Jean-Franois Monin      *)
(* VERIMAG, Universit Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: SortedLasagna.v,v 1.9 2006/07/17 23:29:15 monin Exp $ *)

Require Import NXLayer.
Require Import NEmptXorLayer.
Require Import SortXor.
Require Import SortEmpty.

Require Import Nat2.
Require Import Separable.
Require Import LasagnaTerms.
Require Import Lasagna.

Require Import List.

(* *)

Fixpoint Slistagna_x (e:alte) : sortable :=
  match e with
  | AZ => Sempty
  | ES o => SNElnxor (Slistagna_n o)
  end
with Slistagna_n (o:alto) : sortable :=
  match o with
  | OS e => Snx_term (Slistagna_x e) false
  end.

(*
Definition listagna_x (e:alte) : Set := |Slistagna_x e| .
Definition listagna_n (o:alto) : Set := |Slistagna_n o| .
*)

(*
Definition listagna_cand_x o := n1_list (listagna_n o).
Definition listagna_cand_n e := nx_term (listagna_x e).

Fixpoint lift_listagna_x e1 e2 {struct e1} : listagna_x e1 -> listagna_x (e1 +e e2) :=
  match e1 return listagna_x e1 -> listagna_x (e1 +e e2) with
  | AZ => fun emp => match emp with end
  | ES o1 => map_n1_list (lift_listagna_n o1 e2) 
  end
with lift_listagna_n o1 e2 {struct o1} : listagna_n o1 -> listagna_n (o1 +o e2) :=
  match o1 return listagna_n o1 -> listagna_n (o1 +o e2) with
  | OS e1 => map_nx (lift_listagna_x e1 e2) false
  end.
*)

Fixpoint lift2_listagna_x e {struct e} : |Slistagna_x e| -> |Slistagna_x (ES (OS e))| :=
  match e return |Slistagna_x e| -> |Slistagna_x (ES (OS e))| with
  | AZ => fun emp => match emp with end
  | ES o => map_n1_list (lift2_listagna_n o) 
  end
with lift2_listagna_n o {struct o} : |Slistagna_n o| -> |Slistagna_n (OS (ES o))| :=
  match o return |Slistagna_n o| -> |Slistagna_n (OS (ES o))| with
  | OS e => map_nx (lift2_listagna_x e) false
  end.


(* Normalizing a lasagna *)

Fixpoint listagna_of_lasagna_x (e:alte) : 
  lasagna_x e -> forall b, mayskip_nx (|Slistagna_x e|) b :=
  match e return lasagna_x e -> forall b, mayskip_nx (|Slistagna_x e|) b with
  | AZ => fun e b => MS_no b e
  | ES o => fun x b => 
    match sort (map_xor (listagna_of_lasagna_n o) false x)
       return mayskip_nx (|Slistagna_x (ES o)|) b with
    | nil => MS_no b N1_0
    | z :: nil => MS_yes (nx_forget_false (lift2_listagna_n o z) b)
    | x :: y :: l => MS_no b (N1_2 x y l)
    end
  end
with listagna_of_lasagna_n (o:alto) : lasagna_n o -> |Slistagna_n o| :=
  match o return lasagna_n o -> |Slistagna_n o| with
  | OS e => map_nx_sticky (listagna_of_lasagna_x e) false
  end.

Definition top_listagna_x e := mayskip_nx (|Slistagna_x e|) false.

Definition top_listagna_of_lasagna_x (e:alte) (t:lasagna_x e) : top_listagna_x e :=
  listagna_of_lasagna_x e t false.

Definition listagna_of_parity p : alt_of_parity p -> Set :=
  match p return alt_of_parity p -> Set with
  | Pev => top_listagna_x
  | Pod => fun o => |Slistagna_n o|
  end.

Definition listagna_of_lasagna p : forall (a: alt_of_parity p),
  lasagna_of_parity p a -> listagna_of_parity p a :=
  match p return forall a, lasagna_of_parity p a -> listagna_of_parity p a with
  | Pev => top_listagna_of_lasagna_x
  | Pod => listagna_of_lasagna_n
  end.

(* old version
Fixpoint listagna_of_lasagna_x (e:alte) : lasagna_x e -> listagna_x e :=
  match e return lasagna_x e -> listagna_x e with
  | AZ => fun e => e
  | ES o => fun x => sort (map_xor (listagna_of_lasagna_n o) false x) 
  end
with listagna_of_lasagna_n (o:alto) : lasagna_n o -> listagna_n o :=
  match o return lasagna_n o -> listagna_n o with
  | OS e => map_nx (listagna_of_lasagna_x e) false
  end.
*)

(* Normalizing a term *)

Definition normalized t := listagna_of_parity (parity_of_term t) (alt_of_term t).

Definition normalize_in_list t : normalized t := 
   listagna_of_lasagna (parity_of_term t) (alt_of_term t) (lasagna_of_term t).

(*
Require Import Terms.

Definition t1 := Xor (Hash (Xor Zero (Hash Zero Zero)) Zero) (Hash (Hash Zero Zero) Zero).

Eval compute in (normalize_in_list t1).

*)

(**********************************************************************************)
(* 2nd version : built-in sorted lasagne *)

(*
The form of underlying types is { ln : ... {l1 : {l0 | P0 l0} | P1 l1}...  Pn ln}
but problems vanish because the Pi are singleton predicates
*)

Fixpoint S_sortagna_x (e:alte) : sortable :=
  match e with
  | AZ => Sempty
  | ES o => Sorted_NElnxor (S_sortagna_n o)
  end
with S_sortagna_n (o:alto) : sortable :=
  match o with
  | OS e => Snx_term (S_sortagna_x e) false
  end.

(*
Definition sortagna_x (e:alte) : Set := |S_sortagna_x e| .
Definition sortagna_n (o:alto) : Set := |S_sortagna_n o| .
*)

(* Variable f : sortagna_n o -> sortagna_n (OS (ES o)). *)

Definition mfx e := monot_fun (S_sortagna_x e) (S_sortagna_x (ES (OS e))).
Definition mfn o := monot_fun (S_sortagna_n o) (S_sortagna_n (OS (ES o))).

Definition f0 : |S_sortagna_x AZ| -> |S_sortagna_x (ES (OS AZ))| :=
   fun emp => match emp with end.

Definition mfx0 : mfx AZ.
exists f0; intro x; case x. 
Defined.

Fixpoint lift2_sortagna_x e {struct e} : mfx e :=
  match e return mfx e with
  | AZ => mfx0
  | ES o => s_map_n1 (lift2_sortagna_n o) 
  end
with lift2_sortagna_n o {struct o} : mfn o :=
  match o return mfn o with
  | OS e => s_map_nx (lift2_sortagna_x e) false
  end.

(* Normalizing a lasagna *)

Fixpoint sortagna_of_lasagna_x (e:alte) : 
  lasagna_x e -> forall b, mayskip_nx (|S_sortagna_x e|) b :=
  match e return (lasagna_x e -> forall b : bool, mayskip_nx (|S_sortagna_x e|) b) with
  | AZ => fun emp b => MS_no b emp
  | ES o => fun x b =>
      let (l, sl) := mk_sorted (map_xor (sortagna_of_lasagna_n o) false x) in
      match l return (sorted l -> mayskip_nx (|S_sortagna_x (ES o)|) b) with
      | nil => mk_n1_sorted b N1_0 
      | x :: nil => fun _ =>
              MS_yes (nx_forget_false (app_monot (lift2_sortagna_n o) x) b)
      | x :: y :: l => mk_n1_sorted b (N1_2 x y l)
      end sl
  end
with sortagna_of_lasagna_n (o:alto) : lasagna_n o -> |S_sortagna_n o| :=
  match o return (lasagna_n o -> |S_sortagna_n o|) with
  | OS e => map_nx_sticky (sortagna_of_lasagna_x e) false
  end.

(* warning : Type instead of Set *)
Definition sortagna_of_parity p : alt_of_parity p -> Type :=
  match p return alt_of_parity p -> Type with
  | Pev => fun e => forall b, mayskip_nx (|S_sortagna_x e|) b
  | Pod => fun o => |S_sortagna_n o|
  end.

Definition sortagna_of_lasagna p : 
  forall (a: alt_of_parity p), lasagna_of_parity p a -> sortagna_of_parity p a :=
  match p return forall a, lasagna_of_parity p a -> sortagna_of_parity p a with
  | Pev => sortagna_of_lasagna_x
  | Pod => sortagna_of_lasagna_n
  end.

(* Normalizing a term *)

Definition s_normalized t := sortagna_of_parity (parity_of_term t) (alt_of_term t).

Definition normalize_in_sort t : s_normalized t :=
   sortagna_of_lasagna (parity_of_term t) (alt_of_term t) (lasagna_of_term t).







(* Mise au point
Definition sol_0 : lasagna_x AZ -> forall b, mayskip_nx (sortagna_x AZ) b :=
  fun emp b => MS_no b emp.

sol_x : 
intros sol e. destruct e as [ | o].
  exact sol_0. 
  intros x b; case (mk_sorted (map_xor (sol o) false x)); clear x;
  intro l; case l; clear l. 
    intro sn. unfold sortagna_x; simpl. apply (MS_no). exists (N1_0 (A:=|S_sortagna_n o |)). exact sn. 
    intros x l; case l; clear l.
      intros sxn. unfold sortagna_x; simpl. apply (MS_yes). apply nx_forget_false. 
        generalize (lift2_sortagna_n o). unfold mfn. intros (f, mf).
        exact (f x).  

      intros y l sxyl. unfold sortagna_x; simpl. apply (MS_no). exists (N1_2 x y l). exact sxyl.

sol_x = 
fun (sol : forall o : alto, lasagna_n o -> sortagna_n o) (e : alte) =>
match
  e as a return (lasagna_x a -> forall b : bool, mayskip_nx (sortagna_x a) b)
with
| AZ => sol_0
| ES o =>
    fun (x : lasagna_x (ES o)) (b : bool) =>
    let (l, x0) :=
        mk_sorted (AS:=S_sortagna_n o) (map_xor (sol o) false x) in
    match l as l0 return (sorted l0 -> mayskip_nx (sortagna_x (ES o)) b) with
    | nil =>
        fun sn : sorted nil =>
        MS_no b
          (exist (fun x0 : n1_list (|S_sortagna_n o |) => n1_sorted x0) N1_0
             sn)
    | x0 :: l0 =>
        match
          l0 as l1
          return (sorted (x0 :: l1) -> mayskip_nx (sortagna_x (ES o)) b)
        with
        | nil =>
            fun _ : sorted (x0 :: nil) =>
            MS_yes
              (nx_forget_false
                 match lift2_sortagna_n o with
                 | MF f _ => f x0
                 end b)
        | y :: l1 =>
            fun sxyl : sorted (x0 :: y :: l1) =>
            MS_no b
              (exist (fun x1 : n1_list (|S_sortagna_n o |) => n1_sorted x1)
                 (N1_2 x0 y l1) sxyl)
        end
    end x0
end

*)
